!===========================================================================
!
!  Author: Li, Xiaozhou
!
!  Email: X.Li-2@tudelft.nl or xiaozhouli@live.com
!
!  Creat Time:  2014-10-02
!
!  Last Change: 2014-10-02
!
!  File Name:   dg.f90
!
!  Description: This version, new style, call from python 
!
!===========================================================================
module types
    integer, parameter :: dp = selected_real_kind(15, 307)
end module types

module dg
    use types
    implicit none
    
    !integer, parameter :: dp=kind(0.d0)
    real(dp), parameter :: Pi = 4._dp*atan(1.0_dp)
    ! domain
    real(dp), parameter :: x_left = 0._dp, x_right = 1._dp
    real(dp), parameter :: x_len = x_right - x_left
    ! six points Gaussian quadrature
    real(dp), dimension(6), parameter :: wg = &
        &(/1.7132449237916972e-01_dp/2.0_dp, &
        &  3.6076157304813861e-01_dp/2.0_dp, &
        &  4.6791393457269098e-01_dp/2.0_dp, &
        &  4.6791393457269098e-01_dp/2.0_dp, &
        &  3.6076157304813861e-01_dp/2.0_dp, &
        &  1.7132449237916972e-01_dp/2.0_dp/) 
    real(dp), dimension(6), parameter :: xg = &
        &(/-0.466234757101576013906150777246997304567_dp, &
        &  -0.330604693233132256830699797509952673503_dp, &
        &  -0.119309593041598454315250860840355967709_dp, &
        &   0.119309593041598454315250860840355967709_dp, &
        &   0.330604693233132256830699797509952673503_dp, &
        &   0.466234757101576013906150777246997304567_dp/)

    contains
    !function basis(l, y)
        !! To evaluate the basis polynomial of degree l at the point y
        !integer, intent(in) :: l
        !real(dp), intent(in) :: y 
        !integer :: j
        !real(dp) :: basis

        !basis = P(l,l)
        !do j = l-1, 0, -1
            !basis = basis*y + P(l,j)
        !end do
    !end function basis

    !function basis_x(l, y)
        !! To evaluate the first derivative of the basis polynomial of degre l at the point y
        !integer, intent(in) :: l
        !real(dp), intent(in) :: y
        !integer :: j
        !real(dp) :: basis_x

        !basis_x = P(l,l)*l
        !DO j = l-1, 1, -1
            !basis_x = basis_x*y + P(l,j)*j
        !END DO
    !end function basis_x

    function Legendre(l, y)
        ! To evaluate the Legendre polynomial of degree l at the point y
        integer, intent(in) :: l
        real(dp), intent(in) :: y
        real(dp) :: Legendre

        select case(l)
        case (0)
            Legendre = 1._dp
        case (1)
            Legendre = y
        case (2)
            Legendre = 1.5_dp*y*y - 0.5_dp
        case (3)
            Legendre = y*(-1.5_dp + 2.5_dp*y*y)
        case (4) 
            Legendre = 0.375_dp - y*y*(3.75_dp + 4.375_dp*y*y)
        case default
            Legendre = 0._dp
        end select
    end function Legendre

    function Legendre_x(l, y)
        ! To evaluate the first derivative of the Legendre polynomial of degre l at the point y
        integer, intent(in) :: l
        real(dp), intent(in) :: y
        real(dp) :: Legendre_x

        select case(l)
        case (0)
            Legendre_x = 0._dp
        case (1)
            Legendre_x = 1._dp
        case (2)
            Legendre_x = 3._dp*y
        case (3)
            Legendre_x = -1.5_dp + 7.5_dp*y*y
        case (4) 
            Legendre_x = y*(-7.5_dp + 17.5_dp*y*y)
        case default
            Legendre_x = 0._dp
        END SELECT
    end function Legendre_x

    subroutine init(n, k, x, dx, u, P)
        integer, intent(in) :: n, k
        !real(dp), intent(inout) :: x(n), dx(n), u(0:k,n), P(0:k,0:k)
        real(dp), intent(inout) :: x(:), dx(:), u(:,:), P(:,:)

        integer :: i, j, l
        real(dp) :: h

        h = x_len/n
        dx = h
        do j = 1, n
            x(j) = (j - 0.5_dp)*h
        end do

        do j = 1, n
            do l = 0, k
                u(l+1,j) = 0._dp
                do i = 1, 6
                    u(l+1,j) = u(l+1,j) + wg(i)*sin(2*Pi*(x(j)+dx(j)*xg(i)))* &
                        Legendre(l, xg(i)*2._dp)
                end do
                u(l+1,j) = u(l+1,j)*(2*l+1)
            end do
        end do

        P(1,1) = 1._dp
        P(2,1) = 0._dp
        P(2,2) = 1._dp
        do l = 2, k
            P(l+1, 1) = -P(l-1,1)*(l-1)/l
            do j = 1, l
                P(l+1,j+1) = ((2*l-1)*P(l,j) - (l-1)*P(l-1,j+1))/l
            end do
        end do
    end subroutine init

    subroutine error(n, k, time, x, dx, u)
        integer, intent(in) :: n, k
        !real(dp), intent(in) :: time, x(n), dx(n), u(0:k,n)
        real(dp), intent(in) :: time, x(:), dx(:), u(:,:)
        !real(dp), intent(in) :: u(0:k,n), x(n), dx(n)

        integer :: i, j, l
        real(dp) :: uh, xj, err, err2, erri

        err2 = 0._dp
        erri = 0._dp

        do j = 1, n
            do i = 1, 6
                uh = 0._dp
                xj = x(j) + dx(j)*xg(i)
                do l = 0, k
                    uh = uh + u(l+1,j)*Legendre(l, 2._dp*xg(i))
                end do
                err = abs(sin(2*Pi*(xj-time)) - dg_eval(n, k, x, dx, u, xj))
                !err = abs(sin(2*Pi*(xj-time)) - uh)
                err2 = err2 + dx(j)*wg(i)*err*err
                erri = max(err, erri)
            end do
        end do
        err2 = sqrt(err2/x_len)
        write (*,*) err2, erri
    end subroutine error

    subroutine time_step(n, k, cfl, time, dx, u, P)
        integer, intent(in) :: n, k
        !real(dp), intent(in) :: cfl, time, dx(n), P(0:k,0:k)
        real(dp), intent(in) :: cfl, time, dx(:), P(:,:)
        !real(dp), intent(inout) :: u(0:k,n)
        real(dp), intent(inout) :: u(:,:)

        real(dp) :: t, dt

        dt = x_len/n*cfl 
        t = dt 
        do while (t < time)
            call rk3(n, k, dt, dx, u, P)
            t = t + dt
        end do
        dt = time + dt - t
        call rk3(n, k, dt, dx, u, P)
    end subroutine time_step

    subroutine rk3(n, k, dt, dx, u, P)
        integer, intent(in) :: n, k
        !real(dp), intent(in) :: dt, dx(n), P(0:k,0:k)
        real(dp), intent(in) :: dt, dx(:), P(:,:)
        !real(dp), intent(inout) :: u(0:k,n)
        real(dp), intent(inout) :: u(:,:)
        
        real(dp) :: u_old(k+1,n), res(k+1,n)

        u_old = u
        call residue(n, k, dx, u, P, res)
        u = u_old + dt*res
        call residue(n, k, dx, u, P, res)
        u = 0.75_dp*u_old + 0.25_dp*(u + dt*res)
        call residue(n, k, dx, u, P, res)
        u = (u_old + 2._dp*u + 2._dp*dt*res)/3._dp
    end subroutine rk3

    subroutine residue(n, k, dx, u, P, res)
        integer, intent(in) :: n, k
        !real(dp), intent(in) :: dx(n), u(0:k,n), P(0:k,0:k)
        real(dp), intent(in) :: dx(:), u(:,:), P(:,:)
        real(dp), intent(out) :: res(0:k,n)

        integer :: i, j, l
        real(dp) :: flux(0:n)!, integral

        do j = 1, n
            flux(j) = 0._dp
            do l = 0, k
                flux(j) = flux(j) + u(l+1,j)
            end do
        end do
        flux(0) = flux(n)

        do j = 1, n
            do l = 0, k
                if (mod(l,2) == 0) then
                    res(l,j) = flux(j-1) - flux(j)
                else
                    res(l,j) = -flux(j-1) - flux(j)
                end if
                do i = 0, k
                    res(l,j) = res(l,j) + u(i+1,j)*integral(P, k, i, l)
                end do
                res(l,j) = res(l,j)*(2*l+1)/dx(j)
            end do
        end do
    end subroutine residue

    function integral(P, k, i, j)
        integer, intent(in) :: i, j, k
        !real(dp), intent(in) :: P(0:k,0:k)
        real(dp), intent(in) :: P(:,:)
        real(dp) :: integral

        integer :: m, l

        integral = 0._dp
        do m = 0, i
            do l = 1, j
                if (mod(m+l-1,2) == 0) then
                    integral = integral + P(i+1,m+1)*P(j+1,l+1)*l/(l+m)*2
                end if
            end do
        end do
    end function

    subroutine dg_appr(n, k, cfl, time, x, dx, u, P)
        integer, intent(in) :: n, k
        real(dp), intent(in) :: time, cfl
        !real(dp), intent(inout) :: x(n), dx(n), u(0:k,n), P(0:k,0:k)
        real(dp), intent(inout) :: x(:), dx(:), u(:,:), P(:,:)

        real :: t1, t2

        write (*,*) "Running DG approximation ..."
        call cpu_time(t1)
        call init(n, k, x, dx, u, P)
        call time_step(n, k, cfl, time, dx, u, P)
        call error(n, k, time, x, dx, u)
        call cpu_time(t2)
        write (*,*) "Time: ", t2 - t1
    end subroutine dg_appr

    subroutine dg_eval2(n, k, time, x, dx, u, xj, yj, err)
        integer, intent(in) :: n, k
        real(dp), intent(in) :: time, x(:), dx(:), u(:,:), xj(:)
        real(dp), intent(inout) :: yj(:), err(:)

        !real(dp) :: dg_eval2(size(xj))

        integer :: i, j, l, num
        real(dp) :: x_temp(size(xj))

        x_temp = xj
        num = size(xj)
        !dg_eval2 = 0._dp
        yj = 0._dp
        do i = 1, num
            if (x_temp(i) < x_left) x_temp(i) = x_temp(i) + x_len
            if (x_temp(i) > x_right) x_temp(i) = x_temp(i) - x_len
            do j = 1, n
                if (x_temp(i) <= x(j)+0.5_dp*dx(j)) exit
            end do
            do l = 0, k
                !dg_eval2(i) = dg_eval2(i) + u(l+1,j)*Legendre(l, 2._dp*(xj(i)-x(j))/dx(j))
                yj(i) = yj(i) + u(l+1,j)*Legendre(l, 2._dp*(x_temp(i)-x(j))/dx(j))
            end do
        end do       
        err = sin(2*Pi*(xj-time)) - yj
        !err = abs(sin(2*Pi*(xj-time)) - yj)
    end subroutine dg_eval2

    function dg_eval(n, k, x, dx, u, xj)
        integer, intent(in) :: n, k
        real(dp), intent(in) :: x(:), dx(:), u(:,:), xj
        real(dp) :: dg_eval

        integer :: j, l
        dg_eval = 0._dp
        do j = 1, n
            if (xj < x(j)+0.5_dp*dx(j)) exit
        end do
        do l = 0, k
            dg_eval = dg_eval + u(l+1,j)*Legendre(l, 2._dp*(xj-x(j))/dx(j))
        end do
    end function dg_eval


end module dg
